@@ -2,7 +2,7 @@ from scipy.stats import norm
 from adadmire.mgm import *
 import numpy as np
 import random
-
+from sklearn.metrics.pairwise import nan_euclidean_distances
 
 def pred_continuous(B, Rho, alphap, D_pred, X_pred):
     ind = np.array(D_pred) == 1
@@ -232,40 +232,51 @@ def transform_back(
         X,
         X_scaled
 ):
-    X_back = X_scaled + (X.max(axis=0) - X.min(axis=0)) + X.min(axis=0)
+    X_back = X_scaled * (X.max(axis=0) - X.min(axis=0)) + X.min(axis=0)
     return(X_back)
 
-def rel_dev(x, org):
+def rel_dev(
+        x, 
+        org):
     return(abs((x-org)/org))
 
 def place_anomalies_continuous(
         X,
         n_ano,
-        epsilon
+        epsilon,
+        positive = False
 ):
-    random.seed(123)
+    random.seed(987)
     # first transform data feature-wise to [0,1]
     X_scaled = transform_data(X)
     # Calculate 15% border
     Z = (0.15*X) / (X.max(axis=0) - X.min(axis=0))
-    ano = np.copy(X)
+    ano = [np.copy(X) for i in range(len(epsilon))]
+    change = np.copy(X)
+    dirm = np.copy(X)
     # for each element in X sample shift
     for i in range(X.shape[0]):
         for j in range(X.shape[1]):
             lc = random.uniform( Z[i,j],(Z[i,j]+X_scaled[i,j]))
-            uc = random.uniform( Z[i,j],(1-Z[i,j]-X_scaled[i,j]))
+            uc = random.uniform( Z[i,j],(1+Z[i,j]-X_scaled[i,j]))
             # decide whether lower or upper change
             dir = random.randint(0,1)
             if dir == 1:
-                ano[i,j] = X_scaled[i,j] + epsilon*uc
+                for k in range(len(epsilon)):
+                    ano[k][i,j] = X_scaled[i,j] + epsilon[k]*uc
+                    change[i,j] = uc
+                    dirm[i,j] = 1
             else:
-                ano[i,j] = X_scaled[i,j] - epsilon * lc
+                for k in range(len(epsilon)):
+                    ano[k][i,j] = X_scaled[i,j] - epsilon[k] * lc
+                    change[i,j] = lc
+                    dirm[i,j] = 0
     # transform anomalies back
-    ano_retrans = transform_back(X, ano)
+    ano_retrans = [transform_back(X, tmp) for tmp in ano]
     # place anomalies
     position = np.array([])
     position.shape = (0,2)
-    X_ano = np.copy(X)
+    X_ano = [np.copy(X) for i in range(len(epsilon))]
     k = 0
     while k < n_ano:
         # sample position
@@ -273,10 +284,111 @@ def place_anomalies_continuous(
         col = random.randint(0, (X.shape[1]-1))
         # first check if anomaly already has been placed at that position
         if sum((position == [[row,col]]).all(axis = 1)) == 0:
-            # check if introduced anomaly > 15% deviation (only for epsilon < 1 relevant)
-            if rel_dev(ano_retrans[row,col],X[row,col]) >= 0.15:
-                k = k+1
-                position = np.append(position, [[row, col]], axis = 0)
-                X_ano[row, col] = ano_retrans[row,col]
+            # check if introduced anomaly > 15% deviation (only for epsilon < 1 relevant) and anomaly still positiv 
+            if positive == True:
+                if rel_dev(ano_retrans[0][row,col],X[row,col]) > 0.15 and ano_retrans[(len(epsilon)-1)][row,col] > 0:
+                    k = k+1
+                    position = np.append(position, [[row, col]], axis = 0)
+                    for l in range(len(epsilon)):
+                        X_ano[l][row, col] = ano_retrans[l][row,col]
+            else :
+                if rel_dev(ano_retrans[0][row,col],X[row,col]) > 0.15:
+                    k = k+1
+                    position = np.append(position, [[row, col]], axis = 0)
+                    for l in range(len(epsilon)):
+                        X_ano[l][row, col] = ano_retrans[l][row,col]
     return(X_ano, position)
 
+def impute(
+        X,
+        D,
+        levels, 
+        lam_seq,
+        oIterations = 10000, 
+        oTol = 1e-6 ):
+    # calculate Euclidean distance of all samples to each other
+    # ignore NaN values
+    dist = nan_euclidean_distances(X,X)
+    np.fill_diagonal(dist, np.inf)
+    X_preimp = np.copy(X)
+    D_preimp = np.copy(D)
+
+    for i in range(X.shape[0]):
+        ind_disc = np.where(np.isnan(D[i]))[0]
+        ind_cont = np.where(np.isnan(X[i]))[0]
+        # impute continuous with value of closest sample
+        for j in ind_cont:
+            sample = np.argmin(dist[i])
+            dist_new = np.copy(dist)
+            while np.isnan(X[sample,j]): # check if value for imputation is also nan
+                dist_new[i,sample] = np.inf
+                sample = np.argmin(dist_new[i]) # use second, third, .. closest sample
+            X_preimp[i,j] = X[sample,j]
+        # same for discrete 
+        for j in ind_disc:
+            sample = np.argmin(dist[i])
+            dist_new = np.copy(dist)
+            while np.isnan(D[sample,j]): # check if value for imputation is also nan
+                dist_new[i,sample] = np.inf
+                sample = np.argmin(dist_new[i]) # use second, third, .. closest sample
+            D_preimp[i,j] = D[sample,j]
+
+    # for each lam in lam_seq fit MGM on all data
+    # calculate MSE
+
+    MSE = 2e10
+    MSE_old = 2e10
+    j = 0
+    X_hat = np.zeros(shape=X.shape)
+    D_hat = np.zeros(shape=D.shape)
+    MSE_seq = np.array([])
+    lam_opt = np.array([lam_seq[0]])
+    lam_opt_old = np.array([lam_seq[0]])
+    while ((MSE_old >= MSE) and (j < np.size(lam_seq))):
+        lam_opt_old = np.copy(lam_opt)
+        MSE_old = np.copy(MSE)
+        X_hat_old = np.copy(X_hat)
+        D_hat_old = np.copy(D_hat)
+        lambda_t = np.array([lam_seq[j]])
+        # fit model on all samples
+        Res = Fit_MGM(X_preimp, D_preimp, levels, lambda_t, oIterations, eps=oTol)
+        Res = Res[0]
+        B = Res[0][0]
+        B = B + np.transpose(B)
+        Phi = Res[0][2]
+        Phi = Phi + np.transpose(Phi)
+        alphap = Res[0][3]
+        Rho = Res[0][1]
+        alphaq = Res[0][4]
+        p = B.shape[0]
+        # loop over all samples
+        for i in range(X.shape[0]):
+            X_pred = X_preimp[i]
+            D_pred = D_preimp[i]
+            # predict sample
+            x_hat = pred_continuous(B, Rho, alphap, D_pred, X_pred)
+            # get discrete predictions
+            d_hat = pred_discrete(Rho, X_pred, D_pred, alphaq, Phi, levels,p)
+            levelSum = np.cumsum(levels)
+            levelSum = np.insert(levelSum, 0, 0)
+            d_var = 0
+            d_hat_state = np.copy(D_pred)
+            for k in range(len(d_hat)):
+                if k == levelSum[d_var]:
+                    d_var = d_var + 1
+                    tmp = np.zeros(shape=len(d_hat[levelSum[d_var-1]:levelSum[d_var]]))
+                    tmp[np.where(d_hat[levelSum[d_var-1]:levelSum[d_var]] == max(d_hat[levelSum[d_var-1]:levelSum[d_var]]))] = 1
+                    d_hat_state[levelSum[d_var-1]:levelSum[d_var]] = tmp
+            X_hat[i] = x_hat
+            D_hat[i] = d_hat_state
+        MSE = np.mean((X_hat - X_preimp)**2)
+        MSE_seq = np.append(MSE_seq, MSE)
+        lam_opt = lambda_t
+        j = j+1
+    ind_cont = np.where(np.isnan(X))
+    ind_disc = np.where(np.isnan(D))
+    X_imp = np.copy(X)
+    X_imp[ind_cont] = X_hat_old[ind_cont]
+    D_imp = np.copy(D)
+    D_imp[ind_disc] = D_hat_old[ind_disc]
+    return(X_imp, D_imp, lam_opt_old)
